@@ -1,4 +1,4 @@
-// Copyright (C) 2022-2023 Hibiki AI Limited <info@hibiki-ai.com>
+// Copyright (C) 2022-2024 Hibiki AI Limited <info@hibiki-ai.com>
 //
 // This file is part of nanonext.
 //
@@ -17,6 +17,7 @@
 // nanonext - C level - Mbed TLS Functions -------------------------------------
 
 #define NANONEXT_TLS
+#define NANONEXT_MBED
 #include "nanonext.h"
 
 // utils -----------------------------------------------------------------------
@@ -37,32 +38,32 @@ SEXP rnng_version(void) {
 
 // Statics ---------------------------------------------------------------------
 
-static nano_buf nano_anytobuf(SEXP x) {
+static nano_buf nano_any_buf(const SEXP x, const int strip) {
 
-  nano_buf hash;
+  nano_buf buf;
 
   switch (TYPEOF(x)) {
   case STRSXP:
     if (XLENGTH(x) == 1 && ATTRIB(x) == R_NilValue) {
       const char *s = CHAR(STRING_ELT(x, 0));
-      NANO_INIT(&hash, (unsigned char *) s, strlen(s));
-      return hash;
+      NANO_INIT(&buf, (unsigned char *) s, strlen(s));
+      return buf;
     }
     break;
   case RAWSXP:
     if (ATTRIB(x) == R_NilValue) {
-      NANO_INIT(&hash, (unsigned char *) STDVEC_DATAPTR(x), XLENGTH(x));
-      return hash;
+      NANO_INIT(&buf, (unsigned char *) STDVEC_DATAPTR(x), XLENGTH(x));
+      return buf;
     }
   }
 
-  nano_serialize_xdr(&hash, x);
+  nano_serialize_xdr(&buf, x, strip);
 
-  return hash;
+  return buf;
 
 }
 
-SEXP nano_hashToChar(unsigned char *buf, const size_t sz) {
+static SEXP nano_hash_char(unsigned char *buf, const size_t sz) {
 
   SEXP out;
   char cbuf[sz + sz + 1];
@@ -79,37 +80,33 @@ SEXP nano_hashToChar(unsigned char *buf, const size_t sz) {
 
 }
 
-// SHA-1 and SHA-2 Cryptographic Hash Algorithms -------------------------------
+// SHA-2 Cryptographic Hash Algorithms -----------------------------------------
 
-#define SHA1_KEY_SIZE 20
-#define SHA224_KEY_SIZE 28
-#define SHA256_KEY_SIZE 32
-#define SHA384_KEY_SIZE 48
-#define SHA512_KEY_SIZE 64
+static SEXP nano_sha256(const SEXP x, const SEXP key, const SEXP convert, const int is224) {
 
-SEXP rnng_sha224(SEXP x, SEXP key, SEXP convert) {
-
-  SEXP out;
-  int xc = 0;
+  const size_t outlen = is224 ? SHA224_KEY_SIZE : SHA256_KEY_SIZE;
 #if MBEDTLS_VERSION_MAJOR >= 3
-  unsigned char output[SHA224_KEY_SIZE];
+  unsigned char output[outlen];
 #else
   unsigned char output[SHA256_KEY_SIZE];
 #endif
+  SEXP out;
+  int xc = 0;
 
-  nano_buf xhash = nano_anytobuf(x);
+  nano_buf xhash = nano_any_buf(x, 1);
 
   if (key == R_NilValue) {
 
 #if MBEDTLS_VERSION_MAJOR >= 3
-    xc = mbedtls_sha256(xhash.buf, xhash.cur, output, 1);
+    xc = mbedtls_sha256(xhash.buf, xhash.cur, output, is224);
 #else
-    xc = mbedtls_sha256_ret(xhash.buf, xhash.cur, output, 1);
+    xc = mbedtls_sha256_ret(xhash.buf, xhash.cur, output, is224);
 #endif
+
   } else {
 
-    nano_buf khash = nano_anytobuf(key);
-    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(MBEDTLS_MD_SHA224),
+    nano_buf khash = nano_any_buf(key, 1);
+    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(is224 ? MBEDTLS_MD_SHA224 : MBEDTLS_MD_SHA256),
                          khash.buf, khash.cur, xhash.buf, xhash.cur, output);
     NANO_FREE(khash);
 
@@ -120,36 +117,53 @@ SEXP rnng_sha224(SEXP x, SEXP key, SEXP convert) {
     Rf_error("error generating hash");
 
   if (*NANO_INTEGER(convert)) {
-    out = nano_hashToChar(output, SHA224_KEY_SIZE);
+    out = nano_hash_char(output, outlen);
   } else {
-    out = Rf_allocVector(RAWSXP, SHA224_KEY_SIZE);
-    memcpy(STDVEC_DATAPTR(out), output, SHA224_KEY_SIZE);
+    out = Rf_allocVector(RAWSXP, outlen);
+    memcpy(STDVEC_DATAPTR(out), output, outlen);
   }
 
   return out;
 
 }
 
+SEXP rnng_sha224(SEXP x, SEXP key, SEXP convert) {
+
+  return nano_sha256(x, key, convert, 1);
+
+}
+
 SEXP rnng_sha256(SEXP x, SEXP key, SEXP convert) {
 
+  return nano_sha256(x, key, convert, 0);
+
+}
+
+static SEXP nano_sha512(const SEXP x, const SEXP key, const SEXP convert, const int is384) {
+
+  const size_t outlen = is384 ? SHA384_KEY_SIZE : SHA512_KEY_SIZE;
+#if MBEDTLS_VERSION_MAJOR >= 3
+  unsigned char output[outlen];
+#else
+  unsigned char output[SHA512_KEY_SIZE];
+#endif
   SEXP out;
   int xc = 0;
-  unsigned char output[SHA256_KEY_SIZE];
 
-  nano_buf xhash = nano_anytobuf(x);
+  nano_buf xhash = nano_any_buf(x, 1);
 
   if (key == R_NilValue) {
 
 #if MBEDTLS_VERSION_MAJOR >= 3
-    xc = mbedtls_sha256(xhash.buf, xhash.cur, output, 0);
+    xc = mbedtls_sha512(xhash.buf, xhash.cur, output, is384);
 #else
-    xc = mbedtls_sha256_ret(xhash.buf, xhash.cur, output, 0);
+    xc = mbedtls_sha512_ret(xhash.buf, xhash.cur, output, is384);
 #endif
 
   } else {
 
-    nano_buf khash = nano_anytobuf(key);
-    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(MBEDTLS_MD_SHA256),
+    nano_buf khash = nano_any_buf(key, 1);
+    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(is384 ? MBEDTLS_MD_SHA384 : MBEDTLS_MD_SHA512),
                          khash.buf, khash.cur, xhash.buf, xhash.cur, output);
     NANO_FREE(khash);
 
@@ -160,10 +174,10 @@ SEXP rnng_sha256(SEXP x, SEXP key, SEXP convert) {
     Rf_error("error generating hash");
 
   if (*NANO_INTEGER(convert)) {
-    out = nano_hashToChar(output, SHA256_KEY_SIZE);
+    out = nano_hash_char(output, outlen);
   } else {
-    out = Rf_allocVector(RAWSXP, SHA256_KEY_SIZE);
-    memcpy(STDVEC_DATAPTR(out), output, SHA256_KEY_SIZE);
+    out = Rf_allocVector(RAWSXP, outlen);
+    memcpy(STDVEC_DATAPTR(out), output, outlen);
   }
 
   return out;
@@ -172,125 +186,13 @@ SEXP rnng_sha256(SEXP x, SEXP key, SEXP convert) {
 
 SEXP rnng_sha384(SEXP x, SEXP key, SEXP convert) {
 
-  SEXP out;
-  int xc = 0;
-#if MBEDTLS_VERSION_MAJOR >= 3
-  unsigned char output[SHA384_KEY_SIZE];
-#else
-  unsigned char output[SHA512_KEY_SIZE];
-#endif
-
-  nano_buf xhash = nano_anytobuf(x);
-
-  if (key == R_NilValue) {
-
-#if MBEDTLS_VERSION_MAJOR >= 3
-    xc = mbedtls_sha512(xhash.buf, xhash.cur, output, 1);
-#else
-    xc = mbedtls_sha512_ret(xhash.buf, xhash.cur, output, 1);
-#endif
-
-  } else {
-
-    nano_buf khash = nano_anytobuf(key);
-    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(MBEDTLS_MD_SHA384),
-                         khash.buf, khash.cur, xhash.buf, xhash.cur, output);
-    NANO_FREE(khash);
-
-  }
-
-  NANO_FREE(xhash);
-  if (xc)
-    Rf_error("error generating hash");
-
-  if (*NANO_INTEGER(convert)) {
-    out = nano_hashToChar(output, SHA384_KEY_SIZE);
-  } else {
-    out = Rf_allocVector(RAWSXP, SHA384_KEY_SIZE);
-    memcpy(STDVEC_DATAPTR(out), output, SHA384_KEY_SIZE);
-  }
-
-  return out;
+  return nano_sha512(x, key, convert, 1);
 
 }
 
 SEXP rnng_sha512(SEXP x, SEXP key, SEXP convert) {
 
-  SEXP out;
-  int xc = 0;
-  unsigned char output[SHA512_KEY_SIZE];
-
-  nano_buf xhash = nano_anytobuf(x);
-
-  if (key == R_NilValue) {
-
-#if MBEDTLS_VERSION_MAJOR >= 3
-    xc = mbedtls_sha512(xhash.buf, xhash.cur, output, 0);
-#else
-    xc = mbedtls_sha512_ret(xhash.buf, xhash.cur, output, 0);
-#endif
-
-  } else {
-
-    nano_buf khash = nano_anytobuf(key);
-    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(MBEDTLS_MD_SHA512),
-                         khash.buf, khash.cur, xhash.buf, xhash.cur, output);
-    NANO_FREE(khash);
-
-  }
-
-  NANO_FREE(xhash);
-  if (xc)
-    Rf_error("error generating hash");
-
-  if (*NANO_INTEGER(convert)) {
-    out = nano_hashToChar(output, SHA512_KEY_SIZE);
-  } else {
-    out = Rf_allocVector(RAWSXP, SHA512_KEY_SIZE);
-    memcpy(STDVEC_DATAPTR(out), output, SHA512_KEY_SIZE);
-  }
-
-  return out;
-
-}
-
-SEXP rnng_sha1(SEXP x, SEXP key, SEXP convert) {
-
-  SEXP out;
-  int xc = 0;
-  unsigned char output[SHA1_KEY_SIZE];
-
-  nano_buf xhash = nano_anytobuf(x);
-
-  if (key == R_NilValue) {
-
-#if MBEDTLS_VERSION_MAJOR >= 3
-    xc = mbedtls_sha1(xhash.buf, xhash.cur, output);
-#else
-    xc = mbedtls_sha1_ret(xhash.buf, xhash.cur, output);
-#endif
-
-  } else {
-
-    nano_buf khash = nano_anytobuf(key);
-    xc = mbedtls_md_hmac(mbedtls_md_info_from_type(MBEDTLS_MD_SHA1),
-                         khash.buf, khash.cur, xhash.buf, xhash.cur, output);
-    NANO_FREE(khash);
-
-  }
-
-  NANO_FREE(xhash);
-  if (xc)
-    Rf_error("error generating hash");
-
-  if (*NANO_INTEGER(convert)) {
-    out = nano_hashToChar(output, SHA1_KEY_SIZE);
-  } else {
-    out = Rf_allocVector(RAWSXP, SHA1_KEY_SIZE);
-    memcpy(STDVEC_DATAPTR(out), output, SHA1_KEY_SIZE);
-  }
-
-  return out;
+  return nano_sha512(x, key, convert, 0);
 
 }
 
@@ -302,7 +204,7 @@ SEXP rnng_base64enc(SEXP x, SEXP convert) {
   int xc;
   size_t olen;
 
-  nano_buf hash = nano_anytobuf(x);
+  nano_buf hash = nano_any_buf(x, 0);
   xc = mbedtls_base64_encode(NULL, 0, &olen, hash.buf, hash.cur);
   unsigned char *buf = R_Calloc(olen, unsigned char);
   xc = mbedtls_base64_encode(buf, olen, &olen, hash.buf, hash.cur);
@@ -372,3 +274,51 @@ SEXP rnng_base64dec(SEXP x, SEXP convert) {
   return out;
 
 }
+
+// Mbed TLS Random Data Generator ----------------------------------------------
+
+SEXP rnng_random(SEXP n, SEXP convert) {
+
+  int sz, xc;
+  switch (TYPEOF(n)) {
+  case INTSXP:
+  case LGLSXP:
+    sz = INTEGER(n)[0];
+    if (sz >= 0 && sz <= 1024) break;
+  case REALSXP:
+    sz = Rf_asInteger(n);
+    if (sz >= 0 && sz <= 1024) break;
+  default:
+    Rf_error("'n' must be an integer between 0 and 1024 or coercible to such");
+  }
+
+  SEXP out;
+  unsigned char buf[sz];
+  mbedtls_entropy_context entropy;
+  mbedtls_ctr_drbg_context ctr_drbg;
+  const char *pers = "r-nanonext-rng";
+
+  mbedtls_entropy_init(&entropy);
+  mbedtls_ctr_drbg_init(&ctr_drbg);
+
+  xc = mbedtls_ctr_drbg_seed(&ctr_drbg, mbedtls_entropy_func, &entropy, (const unsigned char *) pers, strlen(pers));
+  if (xc == 0) {
+    xc = mbedtls_ctr_drbg_random(&ctr_drbg, buf, sz);
+  }
+
+  mbedtls_ctr_drbg_free(&ctr_drbg);
+  mbedtls_entropy_free(&entropy);
+
+  if (xc)
+    Rf_error("error generating random bytes");
+
+  if (*NANO_INTEGER(convert)) {
+    out = nano_hash_char(buf, sz);
+  } else {
+    out = Rf_allocVector(RAWSXP, sz);
+    memcpy(STDVEC_DATAPTR(out), buf, sz);
+  }
+
+  return out;
+
+}
